package jve.generated;

import java.util.ArrayList;
import java.util.List;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.JTextComponent;
import javax.swing.text.PlainDocument;

/**
 * This filter acts as a Document for a JTextComponent.  The value of the text field will be
 * used to filter an object based on a given property.  The filter is a simple string contains
 * test.<p>
 * For example entering "Jon" would accept both "Jonathan Brown" and "Don Jones."<p>
 * 
 * Override the accept() method to change the filtering behavior.  The case sensitivity of the
 * filter can be changed via the caseSensitive property.  The default behavior ignores case.<p>
 * 
 * To setup this filter, set the input text field as the filter's visualTextComponent, set the
 * filter as the document of the text field, and set the filteredProperty that will be filtered.
 */
public class SwingPropertyFilter extends PlainDocument implements IFilterBinder {

	protected String filteredProperty;

	protected JTextComponent visualTextComponent;

	protected String value = null;

	protected String lcValue = null;

	protected boolean caseSensitive = false;

	protected List filterListeners = new ArrayList();

	DocumentListener documentListener = new DocumentListener() {
		public void changedUpdate(DocumentEvent e) {
			contentChanged();
		}

		public void insertUpdate(DocumentEvent e) {
			contentChanged();
		}

		public void removeUpdate(DocumentEvent e) {
			contentChanged();
		}
	};

	private String fdebugMsg = null;

	/**
	 * Construct a SwingPropertyFilter.  The filter will reject all canidate objects
	 * until the filteredProperty is set. 
	 *
	 */
	public SwingPropertyFilter() {
		super();
		addDocumentListener(documentListener);
	}

	/**
	 * This is the default filter that is a simple Object's string contains 
	 * the filtered string.
	 * <p> 
	 * If the filteredProperty does not exist on the given object, the object
	 * will be accepted by default.
	 * 
	 * @param o The object to filter upon
	 * @return true if the object is accepted by the filter, false otherwise.
	 */
	public boolean accept(Object o) {
		if (value == null || filteredProperty == null) {
			fdebugMsg = "No value/property to filter on";
			return false;
		}
		// Find out the value of this object's "filteredProperty"
		PropertyHelper h = new PropertyHelper(o, filteredProperty);
		if (h.getValue() == null) {
			fdebugMsg = "can not match property";
			// Do not filter
			return true;
		}
		String oProperty = h.getValue().toString();
		if (!caseSensitive) {
			String p = oProperty.toLowerCase();
			return p.indexOf(lcValue) >= 0;
		}
		return oProperty.indexOf(value) >= 0;
	}

	public void addFilterChangeListener(FilterChangeListener l) {
		if (!filterListeners.contains(l))
			filterListeners.add(l);
	}

	public void removeFilterChangeListener(FilterChangeListener l) {
		filterListeners.remove(l);
	}

	/**
	 * Get the property on the canidate objects that will be filtered based on
	 * the user text input.
	 * @return the filter property
	 */
	public String getFilteredProperty() {
		return filteredProperty;
	}

	/**
	 * Set the property that the canidate objects will be based on.  If the given
	 * property isn't valid on the canidate object, the object will be accepted by default.
	 *  
	 * @param filteredProperty the property to set
	 */
	public void setFilteredProperty(String filteredProperty) {
		String old = getValue();
		this.filteredProperty = filteredProperty;
		fireFilterChanged(this, old, getValue());
	}

	/**
	 * Get the current text value that will be the used to filter the object.
	 * @return the filter value
	 */
	protected String getValue() {
		return value;
	}

	/**
	 * Notify all the registered {@link jve.generated.IFilterBinder.FilterChangeListener FilterChangeListeners} 
	 * that the value of the filter has changed. In most cases, the listeners should 
	 * respond by calling accept on their list of objects.
	 * 
	 * @param src the source filter
	 * @param oldO the old value of the filter
	 * @param newO the new value of the filter
	 */
	protected void fireFilterChanged(Object src, Object oldO, Object newO) {
		FilterChangeEvent evt = new FilterChangeEvent(src, oldO, newO);
		for (int i = 0; i < filterListeners.size(); i++) {
			((FilterChangeListener) filterListeners.get(i)).filterChange(evt);
		}
	}

	/**
	 * Respond to a change of the text field's contents.  This will fire a filter changed event on
	 * the registered {@link jve.generated.IFilterBinder.FilterChangeListener FilterChangeListeners}.
	 *
	 */
	protected void contentChanged() {
		try {
			Object old = getValue();
			value = getContent().getString(0, getLength());
			lcValue = value.toLowerCase();
			fireFilterChanged(this, old, getValue());
			fdebugMsg = null;
		} catch (BadLocationException e) {
			fdebugMsg = e.getMessage();
		}
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		if (fdebugMsg != null) {
			sb.append("***Error**: ");
			sb.append(fdebugMsg);
			sb.append("\n");
		}

		sb.append("SwingPropertyFilter\n\tValue:");
		sb.append(value);
		sb.append("\n");
		return sb.toString();
	}

	/* (non-Javadoc)
	 * @see javax.swing.text.Document#addDocumentListener(javax.swing.event.DocumentListener)
	 */
	public void addDocumentListener(DocumentListener listener) {
		super.addDocumentListener(listener);
		try {
			value = getContent().getString(0, getLength());
			lcValue = value.toLowerCase();
		} catch (BadLocationException e) {
			fdebugMsg = e.getMessage();
		}
	}

	/**
	 * Get whether the filter accept mechanism is sensitive to String case.
	 * @return true if the filter is case sensitive, false otherwise.
	 */
	public boolean isCaseSensitive() {
		return caseSensitive;
	}

	/**
	 * Sets whether the filter should be sensitive to case while filtering.
	 * @param cs true to enable caseSensitivity, false otherwise.
	 */
	public void setCaseSensitive(boolean cs) {
		this.caseSensitive = cs;
	}

	/**
	 * @return Returns the visual text component used by this filter.
	 */
	public JTextComponent getVisualTextComponent() {
		return visualTextComponent;
	}

	/**
	 * Set the visual text component that will be used as input for the filter.
	 * 
	 * @param srcTextComponent The srcTextComponent to set.
	 */
	public void setVisualTextComponent(JTextComponent srcTextComponent) {
		this.visualTextComponent = srcTextComponent;
		if (srcTextComponent != null)
			srcTextComponent.setDocument(this);
	}
}